"""A translation of a Scheme interpreter in Python, modeling the
metacircular evaluator in Ch. 4 of The Structure and Interpretation of
Computer Programs.  (http://www-mitpress.mit.edu/sicp/)

AUTHOR:

    Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

SYNOPSIS:

    It's Scheme.  *grin*

    Scheme is a variant of Lisp, and it's used in many introductory CS
    courses.  It is pretty powerful despite its deceptive size.  For
    more information on Scheme, visit the Scheme home page:

        http://www.swiss.ai.mit.edu/projects/scheme

BUGS:

    One major deficiency of this implementation is that it doesn't
    support tail recursion.  I don't believe I can implement this
    properly without fiddling around with the 'compiler' package and
    primitive bytecodes.  I'll read up on SICP Chapter 5.5 and see if
    this is feasible.

    Another is the broken implementation of pairs --- I use Python's
    list structure to simulate pair structure, but this translation is
    not quite right.  For example, it's a known bug in this
    implementation that:

       (define x 3)
       (define y '(4))
       (eq? (cdr (cons x y) y))

    doesn't work the way it should.

    Another side effect of using Python lists instead of pairs is that
    CONSing two atoms is prohibited.  I don't feel so bad about this
    one, as the ML language doesn't appear to allow this either.

MISCELLANEOUS:

    This source code is for the public domain; I'm writing it just for
    the fun of it.  This is certainly not meant to be efficient or
    even useful.  *grin*

    My future goals is to make this usable enough to run something
    like the explicit-register simulator and compiler in SICP Ch. 5.

"""


from __future__ import nested_scopes

import types, math, sys, pprint, exceptions

"""I need to use the operator module occasionally, but I've also
defined a 'operator' selector function in this program.  To avoid
confusion, I'll rename the 'operator' module as 'python_operator'."""
import operator as python_operator

import string
from symbol import Symbol, true, false, isSymbol
from parser import parse

"""Let's define a personalized SchemeError exception that will be
thrown if bad things happen."""

class SchemeError(Exception): pass


"""We need to save Python's apply() function, since we use it later on
to evaluate primitive procedure calls.  For symmetry, I'm also saving
Python's eval() function within evalInUnderlyingPython()."""
evalInUnderlyingPython = eval
applyInUnderlyingPython = apply



def eval(exp, env):
    """Evaluates an expression 'exp' in an environment 'env'.

    Exercise 4.3 asks us to rewrite this in a more natural
    data-directed manner.
    """
    if isSelfEvaluating(exp):
        return exp
    if isVariable(exp):
        return lookupVariableValue(exp, env)
    if isQuoted(exp):
        return evalQuoted(exp, env)
    if isAssignment(exp):
        return evalAssignment(exp, env)
    if isDefinition(exp):
        return evalDefinition(exp, env)
    if isLet(exp):
        return eval(letToApplication(exp), env)
    if isIf(exp):
        return evalIf(exp, env)
    if isLambda(exp):
        return makeProcedure(lambdaParameters(exp), lambdaBody(exp),
                             env)
    if isBegin(exp):
        return evalSequence(beginActions(exp), env)
    if isCond(exp):
        return eval(condToIf(exp), env)
    if isLoad(exp):
        return load(exp, env)
    if isApplication(exp):
        return apply(eval(operator(exp), env),
                     listOfValues(operands(exp), env))
    raise SchemeError, "Unknown expression type -- eval " + exp


def apply(procedure, arguments):
    """Applies a procedure on a list of arguments."""
    if isPrimitiveProcedure(procedure):
        return applyPrimitiveProcedure(procedure, arguments)
    if isCompoundProcedure(procedure):
        return evalSequence(procedureBody(procedure),
                            extendEnvironment(procedureParameters(procedure),
                                              arguments,
                                              procedureEnvironment(procedure)))
    raise SchemeError, "Unknown procedure type -- apply " + str(procedure)



def listOfValues(exps, env):
    """Given a list of expressions, returns a new list containing the
    values of evaluating each on of them."""
    if isNoOperands(exps): return []
    return [eval(firstOperand(exps), env)] + \
           listOfValues(restOperands(exps), env)

def evalIf(exp, env):
    if isTrue(eval(ifPredicate(exp), env)):
        return eval(ifConsequent(exp), env)
    return eval(ifAlternative(exp), env)


def evalSequence(exps, env):
    if isLastExp(exps): return eval(firstExp(exps), env)
    eval(firstExp(exps), env)
    return evalSequence(restExps(exps), env)


def evalAssignment(exp, env):
    setVariableValue(assignmentVariable(exp),
                     eval(assignmentValue(exp), env),
                     env)
    return Symbol("ok")


def evalDefinition(exp, env):
    defineVariable(definitionVariable(exp),
                   eval(definitionValue(exp), env),
                   env)
    return Symbol("ok")
                     

def isSelfEvaluating(exp):
    if isNumber(exp): return 1
    if isString(exp): return 1
    return 0


def isNumber(x):
    return type(x) in (types.IntType, types.FloatType, types.LongType)


def isString(x):
    return type(x) == types.StringType


def isVariable(exp):
    return isSymbol(exp)


def isQuoted(exp):
    return isTaggedList(exp, Symbol("quote"))

def isUnquoted(exp):
    return isTaggedList(exp, Symbol("unquote"))


def textOfQuotation(exp):
    return exp[1]

def textOfUnquotation(exp):
    return exp[1]

def isTaggedList(exp, tag):
    if isList(exp) and len(exp) > 0:
        return exp[0] == tag
    return 0


def isList(x):
    return type(x) == types.ListType

######################################################################

def evalQuoted(exp, env):
    """Returns a quoted expression, using deepQuotedEval to look for
    UNQUOTE."""
    return deepQuotedEval(textOfQuotation(exp), env)


def deepQuotedEval(exp, env):
    if not isList(exp):
        return exp
    if isNull(exp):
        return exp
    if isUnquoted(exp):
        return eval(textOfUnquotation(exp), env)
    return schemeCons(deepQuotedEval(exp[0], env),
                      deepQuotedEval(exp[1:], env))

######################################################################

def isLet(exp):
    return isTaggedList(exp, Symbol("let"))

def letBindings(exp):
    return exp[1]

def letBindingVariables(bindings):
    if isNull(bindings): return schemeList()
    return schemeCons(bindings[0][0],
                      letBindingVariables(bindings[1:]))

def letBindingValues(bindings):
    if isNull(bindings): return schemeList()
    return schemeCons(bindings[0][1],
                      letBindingValues(bindings[1:]))

def letBody(exp):
    return exp[2:]

def letToApplication(exp):
    bindings = letBindings(exp)
    return schemeCons(makeLambda(letBindingVariables(bindings),
                                 letBody(exp)),
                      letBindingValues(bindings))

######################################################################


def isAssignment(exp):
    """Assignments have the form (set! <var> <value>)."""
    return isTaggedList(exp, Symbol("set!"))


def assignmentVariable(exp):
    return exp[1]


def assignmentValue(exp):
    return exp[2]

######################################################################

def isDefinition(exp):
    return isTaggedList(exp, Symbol("define"))


def definitionVariable(exp):
    if isSymbol(exp[1]):
        return exp[1]
    return exp[1][0]


def definitionValue(exp):
    if isSymbol(exp[1]):
        return exp[2]
    return makeLambda(exp[1][1:], exp[2:])


######################################################################

def isLambda(exp):
    return isTaggedList(exp, Symbol("lambda"))


def lambdaParameters(exp):
    return exp[1]


def lambdaBody(exp):
    return exp[2:]


def makeLambda(parameters, body):
    if parameters.count(Symbol(".")) > 1:
        raise SchemeError, \
              "Parameter list can have at most one vararg parameter -- LAMBDA"
    return [Symbol("lambda"), parameters] + body


######################################################################


def isIf(exp):
    return isTaggedList(exp, Symbol("if"))


def ifPredicate(exp):
    return exp[1]


def ifConsequent(exp):
    return exp[2]


def ifAlternative(exp):
    if not isNull(exp[3:]):
        return exp[3]
    return false


def isNull(x):
    return x == []


def makeIf(predicate, consequent, alternative):
    return [Symbol("if"), predicate, consequent, alternative]


######################################################################

def isBegin(exp):
    return isTaggedList(exp, Symbol("begin"))


def beginActions(exp):
    return exp[1:]


def isLastExp(seq):
    return isNull(seq[1:])


def firstExp(seq):
    return seq[0]


def restExps(seq):
    return seq[1:]


def sequenceToExp(seq):
    if isNull(seq): return seq
    if isLastExp(seq): return firstExp(seq)
    return makeBegin(seq)


def makeBegin(seq):
    return [Symbol("begin")] + seq


######################################################################


def isApplication(exp):
    return isList(exp)


def operator(exp):
    if len(exp) == 0:
        raise SchemeError, \
              "No operator given for procedure application -- OPERATOR"
    return exp[0]


def operands(exp):
    return exp[1:]


def isNoOperands(ops):
    return isNull(ops)


def firstOperand(ops):
    return ops[0]


def restOperands(ops):
    return ops[1:]


######################################################################


def isCond(exp):
    return isTaggedList(exp, Symbol("cond"))


def condClauses(exp):
    return exp[1:]


def isCondElseClause(clause):
    return condPredicate(clause) == Symbol("else")


def condPredicate(clause):
    return clause[0]


def condActions(clause):
    return clause[1:]


def condToIf(exp):
    return expandClauses(condClauses(exp))


def expandClauses(clauses):
    if isNull(clauses):
        return false
    first = clauses[0]
    rest = clauses[1:]
    if isCondElseClause(first):
        if isNull(rest):
            return sequenceToExp(condActions(first))
        raise SchemeError, "else clause isn't last -- condToIf" + clauses
    return makeIf(condPredicate(first),
                  sequenceToExp(condActions(first)),
                  expandClauses(rest))
                                

def isTrue(x):
    return x != false


def isFalse(x):
    return x == false

######################################################################

def isLoad(x):
    return isTaggedList(x, Symbol("load"))

def loadFilename(exp): return exp[1]

def load(exp, env):
    try:
        f = open(loadFilename(exp))
        text = "(begin %s 'ok)" % f.read()
        return eval(parse(text), env)
    except IOError, e:
        raise SchemeError, "LOAD error -- %s" % str(e)
    except TypeError, e:
        raise SchemeError, "LOAD error -- argument must be a string"
        


######################################################################


def makeProcedure(parameters, body, env):
    return [Symbol("procedure"), parameters, body, env]

def isCompoundProcedure(p):
    return isTaggedList(p, Symbol("procedure"))

def procedureParameters(p):
    return p[1]

def procedureBody(p):
    return p[2]

def procedureEnvironment(p):
    return p[3]





######################################################################
"""Environment stuff.  Exercise 4.11 asks us to develop an alternative
environment implementation.

I deviate from the Scheme definitions on LOOKUP-VARIABLE-VALUE,
SET-VARIABLE-VALUE!, and DEFINE-VARIABLE! because of the awkwardness
of using a separate iterative SCAN function --- instead, I've used
Python's iterative control statements."""

def enclosingEnvironment(env):
    return env[1:]

def firstFrame(env):
    return env[0]

THE_EMPTY_ENVIRONMENT = []


def makeFrame(variables, values):
    """Note: here I diverge from SICP's implementation: instead of
    using a cons, I use a Python list.  The selectors frameVariables()
    and frameValues() reflect this."""
    return [variables, values]

def frameVariables(frame):
    return frame[0]

def frameValues(frame):
    return frame[1]

def addBindingToFrame(var, val, frame):
    frame[0] = [var] + frame[0]
    frame[1] = [val] + frame[1]


def extendEnvironment(vars, vals, base_env):
    """Extends an environment with a new set of bindings.  Extended to
    support Scheme's dotted notation for varargs."""
    if Symbol('.') in vars:
        vararg_start = vars.index(Symbol('.'))
        return [makeFrame(vars[:vararg_start] + [vars[vararg_start+1]],
                          vals[:vararg_start] + [vals[vararg_start:]])
                ] + base_env
    elif len(vars) == len(vals):
        return [makeFrame(vars, vals)] + base_env
    elif len(vars) < len(vals):
        raise SchemeError, \
              "Too many arguments supplied %s %s" % (vars, vals)
    raise SchemeError, "Too few arguments supplied %s %s" % (vars, vals)


def lookupVariableValue(var, env):
    while 1:
        if env == THE_EMPTY_ENVIRONMENT:
            raise SchemeError, "Unbound variable " + var
        frame = firstFrame(env)
        vars, vals = frameVariables(frame), frameValues(frame)
        for i in range(len(vars)):
            if vars[i] == var:
                return vals[i]
        env = enclosingEnvironment(env)


def setVariableValue(var, val, env):
    while 1:
        if env == THE_EMPTY_ENVIRONMENT:
            raise SchemeError, "Unbound variable -- SET! " + var
        frame = firstFrame(env)
        vars, values = frameVariables(frame), frameValues(frame)
        for i in range(len(vars)):
            if vars[i] == var:
                values[i] = val
                return
        env = enclosingEnvironment(env)


def defineVariable(var, val, env):
    frame = firstFrame(env)
    vars, values = frameVariables(frame), frameValues(frame)
    for i in range(len(vars)):
        if vars[i] == var:
            values[i] = val
            return
    addBindingToFrame(var, val, frame)


def setupEnvironment():
    initial_environment = extendEnvironment(PRIMITIVE_PROCEDURE_NAMES,
                                            PRIMITIVE_PROCEDURE_OBJECTS,
                                            THE_EMPTY_ENVIRONMENT)
    defineVariable(Symbol("true"), Symbol("true"), initial_environment)
    defineVariable(Symbol("false"), Symbol("false"), initial_environment)
    return initial_environment



######################################################################
######################################################################
"""Here's where we define the primitives of Scheme."""
def isPrimitiveProcedure(proc):
    return isTaggedList(proc, Symbol("primitive"))

def primitiveImplementation(proc): return proc[1]

def schemeCons(head, rest):
    if type(rest) != types.ListType:
        raise SchemeError, "'rest' must be a list."
    return [head] + rest

def schemeCar(L):
    if type(L) != types.ListType:
        raise SchemeError, "CAR --- argument is not a list."
    if len(L) == 0:
        raise SchemeError, "CAR --- cannot take CAR of empty list."
    return L[0]

def schemeCdr(L):
    if type(L) != types.ListType:
        raise SchemeError, "CDR --- argument is not a list."
    if len(L) == 0:
        raise SchemeError, "CDR --- cannot take CDR of empty list."
    return L[1:]

def schemeAppend(*lists):
    if len(lists) == 0:
        return []
    if len(filter(lambda x: type(x) != types.ListType, lists)) > 0:
        raise SchemeError, "APPEND: arguments to APPEND must be lists."
    return reduce(python_operator.add, lists)

def schemeList(*elements):
    return list(elements)

def schemeApply(procedure, arguments):
    return apply(procedure, arguments)

def schemeAdd(*numbers):
    if len(numbers) == 0: return 0
    return reduce(python_operator.add, numbers)

def schemeSubtract(*numbers):
    if len(numbers) == 0:
        raise SchemeError, "primitive '-' --- expects 1 or more arguments"
    return reduce(python_operator.sub, numbers)

def schemeMultiply(*numbers):
    if len(numbers) == 0: return 1
    return reduce(python_operator.mul, numbers)

def schemeDivide(*numbers):
    if len(numbers) == 0:
        raise SchemeError, "primitive '/' --- expects 1 or more arguments"
    return reduce(python_operator.div, numbers)

def schemeRemainder(a, b): return a % b
def schemeQuotient(a, b): return int(a / b)
def schemeSqrt(a): return math.sqrt(a)
def schemeFloor(a): return int(math.floor(a))
def schemeCeiling(a): return int(math.ceil(a))

def schemeNumericalEq(a, b):
    """Here, we need to make sure that we're returning the metacircular
    values of true or false."""
    return schemeBooleanize(a == b)

def schemeLessThan(a, b): return schemeBooleanize(a < b)
def schemeLessThanOrEquals(a, b): return schemeBooleanize(a <= b)
def schemeGreaterThan(a, b): return schemeBooleanize(a > b)
def schemeGreaterThanOrEquals(a, b): return schemeBooleanize(a >= b)

def schemeEqQuestion(a, b):
    if type(a) in (types.IntType, types.FloatType, types.LongType):
        return schemeBooleanize(a == b)
    return schemeBooleanize(a is b)

def schemeEqualQuestion(a, b): return schemeBooleanize(a == b)
def schemeListQuestion(a): return schemeBooleanize(isList(a))
def schemeNullQuestion(a): return schemeBooleanize(isNull(a))

def schemeDisplay(a):
    sys.stdout.write(expressionToString(a))
    return Symbol("ok")
def schemeNewline():
    sys.stdout.write("\n")
    return Symbol("ok")

def schemeStringToSymbol(mystr): return Symbol(mystr)
def schemeSymbolToString(mysymbol): return str(mysymbol)
def schemeNumberToString(mynumber): return str(mynumber)

def schemeStringAppend(*strings):
    if len(strings) == 0: return ""
    return reduce(python_operator.add, strings)

def schemeNot(x):
    if isTrue(x): return false
    return true

def schemeQuit():
    sys.stdout.write("BYE\n")
    sys.exit(0)

def schemeBooleanize(x):
    """Translates what Python considers true and false to the Scheme
    symbols "true" or "false"."""
    if x: return true
    return false

PRIMITIVE_PROCEDURES = [
    [Symbol("car"), schemeCar],
    [Symbol("cdr"), schemeCdr],
    [Symbol("cons"), schemeCons],
    [Symbol("append"), schemeAppend],
    [Symbol("list"), schemeList],
    [Symbol("apply"), schemeApply],
    [Symbol("+"), schemeAdd],
    [Symbol("-"), schemeSubtract],
    [Symbol("*"), schemeMultiply],
    [Symbol("/"), schemeDivide],
    [Symbol("remainder"), schemeRemainder],
    [Symbol("quotient"), schemeQuotient],
    [Symbol("sqrt"), schemeSqrt],
    [Symbol("floor"), schemeFloor],
    [Symbol("ceiling"), schemeCeiling],
    [Symbol("="), schemeNumericalEq],
    [Symbol("<"), schemeLessThan],
    [Symbol("<="), schemeLessThanOrEquals],
    [Symbol(">"), schemeGreaterThan],
    [Symbol(">="), schemeGreaterThanOrEquals],
    [Symbol("eq?"), schemeEqQuestion],
    [Symbol("equal?"), schemeEqualQuestion],
    [Symbol("list?"), schemeListQuestion],
    [Symbol("null?"), schemeNullQuestion],
    [Symbol("display"), schemeDisplay],
    [Symbol("newline"), schemeNewline],
    [Symbol("not"), schemeNot],
    [Symbol("string->symbol"), schemeStringToSymbol],
    [Symbol("symbol->string"), schemeSymbolToString],
    [Symbol("number->string"), schemeNumberToString],
    [Symbol("string-append"), schemeStringAppend],
    [Symbol("quit"), schemeQuit],
    [Symbol("exit"), schemeQuit],
    ]
    
PRIMITIVE_PROCEDURE_NAMES = map(lambda x: x[0], PRIMITIVE_PROCEDURES)

PRIMITIVE_PROCEDURE_OBJECTS = map(lambda proc:
                                  [Symbol("primitive"), proc[1]],
                                  PRIMITIVE_PROCEDURES)

def applyPrimitiveProcedure(proc, args):
    try:
        return applyInUnderlyingPython(primitiveImplementation(proc),
                                       args)
    except Exception, e:
        if isinstance(e, exceptions.SystemExit): raise e
        raise SchemeError, e
######################################################################

def test1():
    """This is a small test harness that tries to at least touch the
    functionality that the interpreter provides.  I should formally
    use unit testing for this when I get the chance."""
    from parser import parse
    global_environment = setupEnvironment()
    commands = parse("""(
    (+ 3 4)
    (define pi 3.1415926)
    pi
    (* 4 2)
    (car (cdr (cons 3 (cons 4 '(5)))))
    ((lambda (x) (* x x)) 42)
    ((lambda (x y z) (+ (* x y) z)) 3 4 5)
    (define square (lambda (x) (* x x)))
    (square 42)
    (define (hypotenuse a b) (sqrt (+ (square a) (square b))))
    (hypotenuse 3 4)
    (define (factorial x)
    (if (= x 0) 1
    (* x (factorial (- x 1)))))
    (factorial 0)
    (factorial 1)
    (factorial 2)
    (factorial 3)
    (begin
    (display 1)
    (display 2)
    (display 3)
    (newline)
    'is-this-working?)
    (define (even? x) (= (remainder x 2) 0))
    (define (odd? x) (not (even? x)))
    (cond ((even? 3) "not-ok")
          ((odd? 2) "still-not-ok")
          (true     "ok"))
    (even? (square 42))
    (odd? (square 42))

    (append '(1 2 3) '(4 5 6))
    (list '(1 2 3) 4)
    '(hello do you see how ,pi is unquoted?)
    '(this (pi (,pi) is deep))
    (define (mylist . elements) elements)
    (append (mylist 1 2 3 4 5)
            (apply mylist '(6 7 8 9 10)))
    )""")
    for c in commands:
        print expressionToString(c)
        print expressionToString(eval(c, global_environment))
        print
    return global_environment


def expressionToString(expr):
    if not isList(expr):
        return str(expr)
    if isCompoundProcedure(expr):
        return expressionToString([Symbol('compound-procedure'),
                                   procedureParameters(expr),
                                   procedureBody(expr)])
    return '(' + string.join(map(expressionToString, expr), ' ') + ')'


def testDriverLoop():
    from parser import parse, ParserError
    from prompt import Prompt
    global_environment = setupEnvironment()
    def promptCallback(s):
        try:
            print expressionToString(eval(parse(s), global_environment))
        except (SchemeError, ParserError), e:
            print "ERROR (%s): %s" % (e.__class__, e)
    p = Prompt(name="PyScheme", callback=promptCallback,
               quit_str=None)
    p.promptLoop()
            

if __name__ == '__main__':
    if len(sys.argv) > 1 and '-t' in sys.argv[1:]:
        test1()
    else:
        print 'Welcome to PyScheme!  Type: (QUIT) to quit.\n'
        testDriverLoop()
